# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing colormap for ~89GHz PMW products."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "pmw_89H"


def call(data_range=[105, 305], cbar_label="TB (K)"):
    """Colormap for displaying ~89GHz PMW data.

    Parameters
    ----------
    data_range : list of float, default=[105, 305]
        * Min and max value for colormap.
        * Ensure the data range matches the range of the algorithm specified for
          use with this colormap
        * This colormap MUST include 105 and 305

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of matplotlib plotting parameters, to ensure consistent
        image output
    """
    min_tb = data_range[0]
    max_tb = data_range[1]

    if min_tb >= 180 or max_tb <= 280:
        raise ("89H TB range MUST include 180 and 280")

    from geoips.image_utils.colormap_utils import create_linear_segmented_colormap

    transition_vals = [
        (min_tb, 180),
        (180, 212),
        (212, 228),
        (228.1, 254),
        (254.1, 280),
        (280, max_tb),
    ]
    transition_colors = [
        ("white", "black"),
        ("#A4641A", "#FC0603"),
        ("#F4CD03", "#F2F403"),
        ("#8CF303", "#0FB503"),
        ("#06DCFD", "#0708B5"),
        ("navy", "white"),
    ]

    # ticks = [xx[0] for xx in transition_vals]

    # special selection of label

    # ticks = [105, 125, 150, 175, 200, 225, 250, 275, 305]
    ticks = [105, 150, 180, 212, 228, 254, 280, 305]

    # selection of min and max values for colormap if needed
    min_tb = transition_vals[0][0]
    max_tb = transition_vals[-1][1]

    LOG.info("Setting cmap")
    mpl_cmap = create_linear_segmented_colormap(
        "pmw_89h", min_tb, max_tb, transition_vals, transition_colors
    )

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_tb, vmax=max_tb)

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = None
    mpl_boundaries = None

    # from geoips.image_utils.mpl_utils import create_colorbar
    # only create colorbar for final imagery
    # cbar = create_colorbar(fig, mpl_cmap, mpl_norm, ticks, cbar_label=cbar_label)
    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": True,
        "cbar_full_width": True,
    }

    # return cbar, min_tb, max_tb
    return mpl_colors_info



#!/bin/env python
"""Update release note index with latest version release reference."""
from sys import argv, exit

index_filename = argv[1]
add_version = argv[2]

# Eventually we will likely support actual versions or latest.  For now only
# support "latest".
# This may end up being just auto-generating the ENTIRE index.rst, which will
# remove the need for specifying a version at all.  For now we are only adding
# the "latest" section to index.rst so the doc build doesn't fail from a non-existent
# reference.  In order to auto-generate the entire index.rst, we would have to
# include the high level minor release version summaries somewhere outside index.rst
# (right now the only specification of the high level summaries is in index.rst,
# so we can't entirely auto-generate it).
# This will be revisited after the 2024 workshop.  For now this will work.
if add_version != "latest":
    print("Actual versions not yet supported - only latest")
    exit(1)
    # Something like this for adding actual version to index.rst
    # Or we may just auto-generate the entire index.rst
    # major, minor, bug_fix = add_version.split(".")
    # minor_version = f"{major}.{minor}"
    # release_note_version = f"v{add_version.replace(".", "_")}"
with open(index_filename, "r") as fobj:
    first = True
    new_lines = []
    last_line = None
    for line in fobj.readlines():
        # Make sure we write all the lines back out
        if "----" not in line or not first:
            if last_line:
                new_lines += [last_line]
        # When we get to the first header ----- line, that means we need to
        # add the new "minor release" section (at the top).
        # If we end up auto-generating the entire index.rst, this will change
        # (since we'll just loop through all the release notes, and write out
        # the sections one by one.  We wouldn't have to keep track of where we
        # were at all).
        if "----" in line and first:
            if add_version == "latest":
                first = False
                new_lines += ["Latest (version on cutting edge of git)\n"]
                new_lines += ["---------------------------------------\n"]
                new_lines += ["\n"]
                new_lines += [".. toctree::\n"]
                new_lines += ["   :maxdepth: 1\n"]
                new_lines += ["\n"]
                new_lines += ["   latest\n"]
                new_lines += ["\n"]
                new_lines += [last_line]
            # Something like this to add actual version to index.rst
            # elif last_line != f"Version {minor_version}":
            #     new_lines += ["Version {minor_version}"]
            #     new_lines += ["-----------------------"]
            #     new_linst += [""]
        last_line = line

    new_lines += [last_line]
# Now write all the lines back out, with the new "latest" section inserted.
with open(index_filename, "w") as fobj:
    fobj.writelines(new_lines)



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Collection of base path names used throughout GeoIPS.

Everything defaults to subdirectories relative to the
REQUIRED environment variable GEOIPS_OUTDIRS.

Individual GEOIPS_OUTDIRS relative paths can be overridden
by setting appropriate environment variables.
"""

# Python Standard Libraries
import logging
from os import getenv
from os.path import exists, dirname, join as pathjoin
import socket

LOG = logging.getLogger(__name__)

PATHS = {}

# Get the base package directory
PATHS["BASE_PATH"] = pathjoin(dirname(__file__), "..")
# URL of the most GeoIPS GitHub hosted documentation.
PATHS["GEOIPS_DOCS_URL"] = r"https://nrlmmd-geoips.github.io/geoips/"

PATHS["GEOIPS_OPERATIONAL_USER"] = False
if getenv("GEOIPS_OPERATIONAL_USER"):
    PATHS["GEOIPS_OPERATIONAL_USER"] = getenv("GEOIPS_OPERATIONAL_USER")

# At a minimum, GEOIPS_OUTDIRS must be defined.
if not getenv("GEOIPS_OUTDIRS"):
    raise KeyError(
        "GEOIPS_OUTDIRS must be set in your environment.  "
        "Please set GEOIPS_OUTDIRS and try again"
    )
else:
    PATHS["GEOIPS_OUTDIRS"] = getenv("GEOIPS_OUTDIRS").rstrip("/")

if getenv("GEOIPS_PACKAGES_DIR") and exists(getenv("GEOIPS_PACKAGES_DIR")):
    PATHS["GEOIPS_PACKAGES_DIR"] = getenv("GEOIPS_PACKAGES_DIR").rstrip("/")
elif getenv("GEOIPS_PACKAGES_DIR"):
    PATHS["GEOIPS_PACKAGES_DIR"] = getenv("GEOIPS_PACKAGES_DIR").rstrip("/")
else:
    PATHS["GEOIPS_PACKAGES_DIR"] = pathjoin(PATHS["BASE_PATH"], "..", "..")

if not getenv("GEOIPS_BASEDIR"):
    PATHS["GEOIPS_BASEDIR"] = pathjoin(PATHS["GEOIPS_PACKAGES_DIR"], "..")
else:
    PATHS["GEOIPS_BASEDIR"] = getenv("GEOIPS_BASEDIR").rstrip("/")

if getenv("GEOIPS_TESTDATA_DIR"):
    PATHS["GEOIPS_TESTDATA_DIR"] = getenv("GEOIPS_TESTDATA_DIR")
else:
    PATHS["GEOIPS_TESTDATA_DIR"] = PATHS["GEOIPS_BASEDIR"] + "/test_data"

if getenv("GEOIPS_DEPENDENCIES_DIR"):
    PATHS["GEOIPS_DEPENDENCIES_DIR"] = getenv("GEOIPS_DEPENDENCIES_DIR")
else:
    PATHS["GEOIPS_DEPENDENCIES_DIR"] = PATHS["GEOIPS_BASEDIR"] + "/geoips_dependencies"

# Location for writing out presectored data files, but unregistered
if getenv("PRESECTORED_DATA_PATH"):
    PATHS["PRESECTORED_DATA_PATH"] = getenv("PRESECTORED_DATA_PATH").rstrip("/")
else:
    PATHS["PRESECTORED_DATA_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "sectored"
    )

# Location for writing out preread, but unsectored/registered, data files
if getenv("PREREAD_DATA_PATH"):
    PATHS["PREREAD_DATA_PATH"] = getenv("PREREAD_DATA_PATH").rstrip("/")
else:
    PATHS["PREREAD_DATA_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "unsectored"
    )

# Location for writing out preregistered data files, but no algorithms applied
if getenv("PREREGISTERED_DATA_PATH"):
    PATHS["PREREGISTERED_DATA_PATH"] = getenv("PREREGISTERED_DATA_PATH").rstrip("/")
else:
    PATHS["PREREGISTERED_DATA_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "registered"
    )

# Location for writing out precalculated data files (algorithms applied)
if getenv("PRECALCULATED_DATA_PATH"):
    PATHS["PRECALCULATED_DATA_PATH"] = getenv("PRECALCULATED_DATA_PATH").rstrip("/")
else:
    PATHS["PRECALCULATED_DATA_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "algorithms"
    )

# Location for writing out pregenerated "clean" imagery files
if getenv("CLEAN_IMAGERY_PATH"):
    PATHS["CLEAN_IMAGERY_PATH"] = getenv("CLEAN_IMAGERY_PATH").rstrip("/")
else:
    PATHS["CLEAN_IMAGERY_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "clean_imagery"
    )

# Location for writing out pregenerated "clean" imagery files
if getenv("ANNOTATED_IMAGERY_PATH"):
    PATHS["ANNOTATED_IMAGERY_PATH"] = getenv("ANNOTATED_IMAGERY_PATH").rstrip("/")
else:
    PATHS["ANNOTATED_IMAGERY_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "annotated_imagery"
    )

# Location for writing out precalculated data files (algorithms applied)
if getenv("FINAL_DATA_PATH"):
    PATHS["FINAL_DATA_PATH"] = getenv("FINAL_DATA_PATH").rstrip("/")
else:
    PATHS["FINAL_DATA_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "final"
    )

# Location for writing out pregenerated geolocation netcdf files
if getenv("PREGENERATED_GEOLOCATION_PATH"):
    PATHS["PREGENERATED_GEOLOCATION_PATH"] = getenv(
        "PREGENERATED_GEOLOCATION_PATH"
    ).rstrip("/")
else:
    PATHS["PREGENERATED_GEOLOCATION_PATH"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "geolocation"
    )

# GEOIPS_COPYRIGHT determines what organization name displays in imagery titles, etc.
PATHS["GEOIPS_COPYRIGHT"] = "NRL-Monterey"
if getenv("GEOIPS_COPYRIGHT"):
    PATHS["GEOIPS_COPYRIGHT"] = getenv("GEOIPS_COPYRIGHT")

# GEOIPS_COPYRIGHT_ABBREVIATED provides an abbreviated version of the
# copyright, best for filenames, etc.
PATHS["GEOIPS_COPYRIGHT_ABBREVIATED"] = "NRLMRY"
if getenv("GEOIPS_COPYRIGHT_ABBREVIATED"):
    PATHS["GEOIPS_COPYRIGHT_ABBREVIATED"] = getenv("GEOIPS_COPYRIGHT_ABBREVIATED")

PATHS["GEOIPS_RCFILE"] = ""
if getenv("GEOIPS_RCFILE"):
    PATHS["GEOIPS_RCFILE"] = getenv("GEOIPS_RCFILE")

PATHS["TC_TEMPLATE"] = pathjoin(
    PATHS["BASE_PATH"], "plugins", "yaml", "sectors", "dynamic", "tc_web_template.yaml"
)
if getenv("TC_TEMPLATE"):
    PATHS["TC_TEMPLATE"] = getenv("TC_TEMPLATE")

PATHS["DEFAULT_QUEUE"] = None
if getenv("DEFAULT_QUEUE"):
    PATHS["DEFAULT_QUEUE"] = getenv("DEFAULT_QUEUE")

PATHS["BOXNAME"] = socket.gethostname()
if not getenv("HOME"):
    # Windows
    PATHS["HOME"] = getenv("HOMEDRIVE") + getenv("HOMEPATH")
else:
    PATHS["HOME"] = getenv("HOME").rstrip("/")

if not getenv("SCRATCH"):
    PATHS["SCRATCH"] = pathjoin(getenv("GEOIPS_OUTDIRS"), "scratch")
else:
    PATHS["SCRATCH"] = getenv("SCRATCH").rstrip("/")

if not getenv("LOCALSCRATCH"):
    PATHS["LOCALSCRATCH"] = PATHS["SCRATCH"]
else:
    PATHS["LOCALSCRATCH"] = getenv("LOCALSCRATCH").rstrip("/")

if not getenv("SHAREDSCRATCH"):
    PATHS["SHAREDSCRATCH"] = PATHS["SCRATCH"]
else:
    PATHS["SHAREDSCRATCH"] = getenv("SHAREDSCRATCH").rstrip("/")

if not getenv("GEOIPS_ANCILDAT"):
    PATHS["GEOIPS_ANCILDAT"] = pathjoin(PATHS["GEOIPS_OUTDIRS"], "ancildat")
elif getenv("GEOIPS_ANCILDAT"):
    PATHS["GEOIPS_ANCILDAT"] = getenv("GEOIPS_ANCILDAT").rstrip("/")
# Separating the auto-generated files from the source files allows for
# individual users to read from the shared ancildat, and write to their own
# auto-generated location
if not getenv("GEOIPS_ANCILDAT_AUTOGEN"):
    PATHS["GEOIPS_ANCILDAT_AUTOGEN"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "ancildat_autogen"
    )
elif getenv("GEOIPS_ANCILDAT_AUTOGEN"):
    PATHS["GEOIPS_ANCILDAT_AUTOGEN"] = getenv("GEOIPS_ANCILDAT_AUTOGEN").rstrip("/")

if not getenv("LOGDIR"):
    PATHS["LOGDIR"] = pathjoin(PATHS["GEOIPS_OUTDIRS"], "logs")
else:
    PATHS["LOGDIR"] = getenv("LOGDIR").rstrip("/")

if not getenv("GEOIPSDATA"):
    PATHS["GEOIPSDATA"] = pathjoin(PATHS["GEOIPS_OUTDIRS"], "geoipsdata")
else:
    PATHS["GEOIPSDATA"] = getenv("GEOIPSDATA").rstrip("/")

if not getenv("GEOIPS_VERS"):
    PATHS["GEOIPS_VERS"] = "0.0.0"
else:
    PATHS["GEOIPS_VERS"] = getenv("GEOIPS_VERS")


if getenv("TCWWW"):
    PATHS["TCWWW"] = getenv("TCWWW").rstrip("/")
else:
    PATHS["TCWWW"] = pathjoin(PATHS["GEOIPS_OUTDIRS"], "preprocessed", "tcwww")

if getenv("TCWWW_URL"):
    PATHS["TCWWW_URL"] = getenv("TCWWW_URL").rstrip("/")
else:
    PATHS["TCWWW_URL"] = PATHS["TCWWW"]

if getenv("PUBLICWWW"):
    PATHS["PUBLICWWW"] = getenv("PUBLICWWW").rstrip("/")
else:
    PATHS["PUBLICWWW"] = pathjoin(PATHS["GEOIPS_OUTDIRS"], "preprocessed", "publicwww")

if getenv("PUBLICWWW_URL"):
    PATHS["PUBLICWWW_URL"] = getenv("PUBLICWWW_URL").rstrip("/")
else:
    PATHS["PUBLICWWW_URL"] = PATHS["PUBLICWWW"]

if getenv("PRIVATEWWW"):
    PATHS["PRIVATEWWW"] = getenv("PRIVATEWWW").rstrip("/")
else:
    PATHS["PRIVATEWWW"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "preprocessed", "privatewww"
    )

if getenv("PRIVATEWWW_URL"):
    PATHS["PRIVATEWWW_URL"] = getenv("PRIVATEWWW_URL").rstrip("/")
else:
    PATHS["PRIVATEWWW_URL"] = PATHS["PRIVATEWWW"]

PATHS["TC_DECKS_DB"] = pathjoin(
    PATHS["GEOIPS_OUTDIRS"], "longterm_files", "tc", "tc_decks.db"
)
if getenv("TC_DECKS_DB"):
    PATHS["TC_DECKS_DB"] = getenv("TC_DECKS_DB")

if getenv("TC_DECKS_DIR"):
    PATHS["TC_DECKS_DIR"] = getenv("TC_DECKS_DIR").rstrip("/")
else:
    PATHS["TC_DECKS_DIR"] = pathjoin(
        PATHS["GEOIPS_OUTDIRS"], "longterm_files", "tc", "decks"
    )


def make_dirs(path):
    """Make directories, catching exceptions if directory already exists.

    Parameters
    ----------
    path : str
        Path to directory to create

    Returns
    -------
    str
        Path if successfully created
    """
    from os import makedirs

    if not exists(path):
        try:
            LOG.info("Creating directory %s", path)
            makedirs(path, mode=0o755)
        except OSError as resp:
            LOG.warning(
                "%s: We thought %s did not exist, but then it did. "
                "Not trying to make directory",
                resp,
                path,
            )
    return path



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""TestPluginRegistry Class used for Unit Testing the Plugin Registries."""

from glob import glob
from importlib import import_module, metadata
import logging
from os.path import basename, splitext
from pprint import pformat
import pytest
import json
import yaml

from geoips.plugin_registry import PluginRegistry
from geoips.errors import PluginRegistryError

LOG = logging.getLogger(__name__)


class PluginRegistryValidator(PluginRegistry):
    """Subclass of PluginRegistry which adds functionality for unit testing."""

    def __init__(self, fpaths=None):
        """Initialize TestPluginRegistry Class."""
        super().__init__(fpaths)

    def validate_plugin_types_exist(self, reg_dict, reg_path):
        """Test that all top level plugin types exist in each registry file."""
        expected_plugin_types = ["yaml_based", "module_based", "text_based"]
        for p_type in expected_plugin_types:
            if p_type not in reg_dict:
                error_str = f"Expected plugin type '{p_type}' to be in the registry but"
                error_str += f" wasn't. This was in file '{reg_path}'."
                raise PluginRegistryError(error_str)

    def validate_all_registries(self):
        """Validate all registries in the current installation.

        This should be run during testing, but not at runtime.
        Ensure we do not fail catastrophically for a single bad plugin
        at runtime, so test up front to test validity.
        """
        for reg_path in self.registry_files:
            pkg_plugins = json.load(open(reg_path, "r"))
            self.validate_registry(pkg_plugins, reg_path)

    def validate_registry(self, current_registry, fpath):
        """Test all plugins found in registered plugins for their validity."""
        try:
            self.validate_plugin_types_exist(current_registry, fpath)
        except PluginRegistryError as e:
            # xfail if this is a test, otherwise just raise PluginRegistryError
            if self._is_test:
                pytest.xfail(str(e))
            else:
                raise PluginRegistryError(e)
        try:
            self.validate_registry_interfaces(current_registry)
        except PluginRegistryError as e:
            # xfail if this is a test, otherwise just raise PluginRegistryError
            if self._is_test:
                pytest.xfail(str(e))
            else:
                raise PluginRegistryError(e)
        for plugin_type in current_registry:
            for interface in current_registry[plugin_type]:
                for plugin in current_registry[plugin_type][interface]:
                    try:
                        if interface == "products":
                            for subplg in current_registry[plugin_type][interface][
                                plugin
                            ]:
                                self.validate_plugin_attrs(
                                    plugin_type,
                                    interface,
                                    (plugin, subplg),
                                    current_registry[plugin_type][interface][plugin][
                                        subplg
                                    ],
                                )
                        elif plugin_type != "text_based":
                            self.validate_plugin_attrs(
                                plugin_type,
                                interface,
                                plugin,
                                current_registry[plugin_type][interface][plugin],
                            )
                    except PluginRegistryError as e:
                        # xfail if this is a test,
                        # otherwise just raise PluginRegistryError
                        if self._is_test:
                            pytest.xfail(str(e))
                        else:
                            raise PluginRegistryError(e)

    def validate_plugin_attrs(self, plugin_type, interface, name, plugin):
        """Test non-product plugin for all required attributes."""
        missing = []
        if plugin_type == "yaml_based" and interface != "products":
            attrs = [
                "docstring",
                "family",
                "interface",
                "package",
                "plugin_type",
                "relpath",
            ]
        elif plugin_type == "yaml_based":
            attrs = [
                "docstring",
                "family",
                "interface",
                "package",
                "plugin_type",
                "product_defaults",
                "source_names",
                "relpath",
            ]
        else:
            attrs = [
                "docstring",
                "family",
                "interface",
                "package",
                "plugin_type",
                "signature",
                "relpath",
            ]
        for attr in attrs:
            try:
                plugin[attr]
            except KeyError:
                missing.append(attr)
        if missing:
            raise PluginRegistryError(
                f"Plugin '{name}' is missing the following required "
                f"top-level properties: '{missing}'"
            )

    def validate_registry_interfaces(self, current_registry):
        """Test Plugin Registry interfaces validity."""
        yaml_interfaces = []
        module_interfaces = []

        # NOTE: all <pkg>/interfaces/__init__.py files MUST
        # contain a "module_based_interfaces" list and a
        # "yaml_based_interfaces" list - this is how we
        # identify the valid interface names.
        # We must avoid actually importing the interfaces within
        # the geoips repo - or we will end up with a circular import
        # due to BaseInterface.
        for pkg in metadata.entry_points(group="geoips.plugin_packages"):
            try:
                mod = import_module(f"{pkg.value}.interfaces")
            except ModuleNotFoundError as resp:
                if f"No module named '{pkg.value}.interfaces'" in str(resp):
                    continue
                else:
                    raise ModuleNotFoundError(resp)
            module_interfaces += mod.module_based_interfaces
            yaml_interfaces += mod.yaml_based_interfaces

        bad_interfaces = []
        for plugin_type in ["module_based", "yaml_based"]:
            for interface in current_registry[plugin_type]:
                if (
                    interface not in module_interfaces
                    and interface not in yaml_interfaces
                ):
                    error_str = f"\nPlugin type '{plugin_type}' does not allow "
                    error_str += f"interface '{interface}'.\n"
                    error_str += "\nValid interfaces: "
                    if plugin_type == "module_based":
                        interface_list = module_interfaces
                    else:
                        interface_list = yaml_interfaces
                    error_str += f"\n{interface_list}\n"
                    error_str += "\nPlease update the following plugins "
                    error_str += "to use a valid interface:\n"
                    error_str += pformat(current_registry[plugin_type][interface])
                    bad_interfaces.append(error_str)
        if bad_interfaces:
            error_str = "The following interfaces were not valid:\n"
            for error in bad_interfaces:
                error_str += error
            raise PluginRegistryError(error_str)


class TestPluginRegistry:
    """
    Pytest-based Unit Test for the PluginRegistry Class.

    Note: Since we are not able to initlialize this class due to restrictions placed by
    Pytest, if you want to change the test files used, simply replace the location below
    with the location of your new test files.
    """

    default_fpaths = glob(
        str(__file__).replace("test_plugin_registries.py", "files/**/*.yaml"),
        recursive=True,
    )

    pr_validator = PluginRegistryValidator(default_fpaths)

    # Couldn't implement this class via inheritance because PyTest Raised this error:
    # PytestCollectionWarning: cannot collect test class 'TestPluginRegistry' because
    # it has a __init__ constructor (from: test_plugin_registries.py)

    # This isn't a problem though, as we can just set it as a class attribute and move
    # forward as usual. Pytest is just picky about having __init__ in Test Classes.

    # def __init__(self, fpaths=default_fpaths):
    #     super().__init__(fpaths)

    def generate_id(fpath):
        """Generate pytest id for current test."""
        test_id = "bad"
        if "/good/" in fpath:
            test_id = "good"
        return f"{test_id}-{splitext(basename(fpath))[0]}"

    def test_registered_plugin_property(self):
        """Ensure registered_plugins is valid in its nature."""
        print(self.pr_validator.registered_plugins)
        assert isinstance(self.pr_validator.registered_plugins, dict)
        assert "yaml_based" in self.pr_validator.registered_plugins
        assert "module_based" in self.pr_validator.registered_plugins
        assert "text_based" in self.pr_validator.registered_plugins

    def test_interface_mapping_property(self):
        """Ensure interface_mapping is valid in its nature."""
        print(self.pr_validator.interface_mapping)
        assert isinstance(self.pr_validator.interface_mapping, dict)
        assert "yaml_based" in self.pr_validator.interface_mapping
        assert "module_based" in self.pr_validator.interface_mapping
        assert "text_based" in self.pr_validator.interface_mapping
        assert isinstance(self.pr_validator.interface_mapping["yaml_based"], list)
        assert isinstance(self.pr_validator.interface_mapping["module_based"], list)
        assert isinstance(self.pr_validator.interface_mapping["text_based"], list)

    @pytest.mark.parametrize("fpath", pr_validator.registry_files, ids=generate_id)
    def test_all_registries(self, fpath):
        """Test all available yaml registries."""
        current_registry = yaml.safe_load(open(fpath, "r"))
        self.pr_validator.validate_registry(current_registry, fpath)



"""VIIRS SDR Satpy reader.

This VIIRS reader is designed for reading the NPP/JPSS SDR HDF5 files.
The input files are produced by CSPP Polar (CSPP RDR pipeline),
and the read by satpy.

V1.1.0:  NRL-Monterey, Aug. 2024

"""

# Python Standard Libraries
import logging
import os

# Installed Libraries
import xarray as xr
import numpy as np
import h5py
from pandas import date_range
from pykdtree.kdtree import KDTree

# If this reader is not installed on the system, don't fail altogether, just skip this
# import. This reader will not work if the import fails, and the package will have to be
# installed to process data of this type.
LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "viirs_sdr_hdf5"

try:
    import satpy
except ImportError:
    LOG.info("Failed import satpy. If you need it, install it.")

VARLIST = {
    "DNB": ["DNB"],
    "IMG": ["I04", "I05"],
    "IMG-Vis": ["I01", "I02", "I03"],
    "MOD": ["M07", "M08", "M10", "M11", "M12", "M13", "M14", "M15", "M16"],
    "MOD-Vis": ["M01", "M02", "M03", "M04", "M05", "M06", "M09"],
}


def bowtie_correction(band, lat, lon):
    """Correct input data for the instrument bowtie effect.

    Correction derived from: Remote Sens. 2016, 8, 79.
    """
    # Unfold
    ord_lat = np.sort(lat, axis=0)
    unfold_idx = np.argsort(lat, axis=0)
    # no_shift_flag = np.argsort(ord_lat, axis=0) == unfold_idx

    rad_fold = np.take_along_axis(band, unfold_idx, axis=0)
    sort_lon = np.take_along_axis(lon, unfold_idx, axis=0)

    if np.all(np.isnan(rad_fold)):
        LOG.debug("All nan band, no bowtie correction")
        return rad_fold, ord_lat, sort_lon

    # Adjust lon, not used for satpy
    # Only need for manual read with overlapping granuales
    # ord_lon = np.empty(lon.shape)
    # xi = np.arange(sort_lon.shape[0])

    # for x in range(lon.shape[1]):
    # 0 shift xi values
    # xo = xi[no_shift_flag[:, x]]

    # if all(no_shift_flag[:, x]):
    # if there was no shift in the column
    # ord_lon[:, x] = lon[:, x]
    # continue
    # elif not any(no_shift_flag[:, x]):
    # if the whole column was shifted (should be rare)
    # ord_lon[:, x] = sort_lon[:, x]
    # continue

    # longitude values that were not shifted
    # noshift_lon = sort_lon[xo,x]
    # noshift_lon = sort_lon[:, x][no_shift_flag[:, x]]
    # replace only values that were shifted
    # nsf = no_shift_flag[:, x]

    # ord_lon[nsf, x] = noshift_lon
    # ord_lon[~nsf, x] = np.interp(xi, xo, noshift_lon)[~nsf]

    # Resample
    point_mask = np.isnan(rad_fold)

    good_points = np.dstack((ord_lat[~point_mask], sort_lon[~point_mask]))[0]
    bad_points = np.dstack((ord_lat[point_mask], sort_lon[point_mask]))[0]

    res_band = rad_fold.copy()
    good_rad = rad_fold[~point_mask]
    rad_idx = np.indices(rad_fold.shape)
    ridx, ridy = rad_idx[0][point_mask], rad_idx[1][point_mask]
    os.environ["OMP_NUM_THREADS"] = "64"

    kd_tree = KDTree(good_points)
    # print("Querying")
    dist, idx = kd_tree.query(bad_points, k=4)  # ,workers=4)

    for i in range(bad_points.shape[0]):
        xi, yi = ridx[i], ridy[i]

        if np.any(dist[i] == 0):
            # weight the zero to a small value
            weight = np.where(dist[i] == 0, 1e-6, dist[i])
            res_band[xi, yi] = np.average(good_rad[idx[i]], weights=1 / weight)
            continue

        res_band[xi, yi] = np.average(good_rad[idx[i]], weights=1 / dist[i])

    return res_band, ord_lat.astype(np.float64), sort_lon.astype(np.float64)


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read VIIRS SDR hdf5 data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.
        * Conforms to geoips xarray standards, see more in geoips documentation.
    """
    # print("Reading")
    tmp_scn = satpy.Scene(reader="viirs_sdr", filenames=fnames)
    scn_start, scn_end = tmp_scn.start_time, tmp_scn.end_time
    base_fnames = list(map(os.path.basename, fnames))

    full_xr = {}
    if metadata_only:
        # average resolution
        # sensor, plat name
        tmp_scn.load([tmp_scn.available_dataset_names()[0]])
        tmp_attrs = tmp_scn[tmp_scn.available_dataset_names()[0]].attrs
        tmp_xr = xr.Dataset(
            attrs={
                "source_file_name": base_fnames[0],
                "start_datetime": scn_start,
                "end_datetime": scn_end,
                "source_name": tmp_attrs["sensor"],
                "platform_name": tmp_attrs["platform_name"],
                "data_provider": "NOAA",
                "sample_distance_km": 1,
                "interpolation_radius_of_influence": 1000,  # guess!
            }
        )
        tmp_dict = {"METADATA": tmp_xr}
        return tmp_dict

    # trim VARLIST based on channels requested
    if chans:
        tmp_vl = VARLIST.copy()
        for key, val in tmp_vl.items():
            km = [c[:3] in val for c in chans]
            if not any(km):
                VARLIST.pop(key)
            else:
                matches = [s for s in val for c in chans if s in c]
                VARLIST[key] = matches

    # could opimize more
    tmp_coor = {}
    for var in VARLIST:
        tmp_dask = {}
        dataset_ids = [
            idx
            for idx in tmp_scn.available_dataset_ids()
            if idx["name"] in VARLIST[var]
        ]
        if len(dataset_ids) == 0:
            # print("No datasets found for {}.".format(VARLIST[var]))
            continue

        for d in dataset_ids:
            # print("Loading {}".format(d))
            tmp_scn.load([d])
            full_key = tmp_scn[d].attrs["name"] + tmp_scn[d].attrs[
                "calibration"
            ].capitalize()[:3].replace("Bri", "BT")

            tmp_ma = tmp_scn[d].to_masked_array().data

            #
            tmp_scn.load([d])

            lat = tmp_scn[d].area.lats.to_masked_array().data
            lon = tmp_scn[d].area.lons.to_masked_array().data

            # bowtie correction
            band_data, band_lat, band_lon = bowtie_correction(tmp_ma, lat, lon)

            tmp_dask |= {full_key: (("dim_0", "dim_1"), band_data)}

        # coordinates
        tmp_coor["latitude"] = (
            ("dim_0", "dim_1"),
            band_lat,
        )
        tmp_coor["longitude"] = (
            ("dim_0", "dim_1"),
            band_lon,
        )
        # print("Setting cal vals")
        # sample time to the proper shape (N*48), while lat/lon are ()
        time_range = date_range(
            start=scn_start, end=scn_end, periods=tmp_coor["latitude"][1].shape[0]
        ).values
        interp_time = np.tile(time_range, (tmp_coor["latitude"][1].shape[1], 1)).T
        tmp_coor["time"] = (("dim_0", "dim_1"), interp_time)
        # # print(tmp_coor["latitude"][1].shape)
        # raise

        tmp_attrs = tmp_scn[VARLIST[var][0]].attrs

        cal_params = [
            "satellite_azimuth_angle",
            "satellite_zenith_angle",
            "solar_azimuth_angle",
            "solar_zenith_angle",
        ]

        if var == "DNB":
            cal_params = [
                "dnb_lunar_azimuth_angle",
                "dnb_lunar_zenith_angle",
                "dnb_satellite_azimuth_angle",
                "dnb_satellite_zenith_angle",
                "dnb_solar_azimuth_angle",
                "dnb_solar_zenith_angle",
            ]

        tmp_scn.load(cal_params)
        tmp_cal_params = {
            i.removeprefix("dnb_"): (("dim_0", "dim_1"), tmp_scn[i].to_masked_array())
            for i in cal_params
        }

        if var == "DNB":
            try:
                from lunarref.lib.liblunarref import lunarref

                # tmp_scn.load(["dnb_moon_illumination_fraction"])
                # this results in the wrong value..
                # np.arccos((tmp_scn["dnb_moon_illumination_fraction"].data/50)-1)

                dnb_geofile = [i for i in fnames if "GDNBO" in os.path.basename(i)][0]
                h5_dnb = h5py.File(dnb_geofile)
                phase_ang = h5_dnb["All_Data/VIIRS-DNB-GEO_All/MoonPhaseAngle"][...]

                lunarref_data = lunarref(
                    tmp_dask["DNBRad"][1],
                    tmp_cal_params["solar_zenith_angle"][1],
                    tmp_cal_params["lunar_zenith_angle"][1],
                    scn_start.strftime("%Y%m%d%H"),
                    scn_start.strftime("%M"),
                    phase_ang,
                )
                lunarref_data = np.ma.masked_less_equal(lunarref_data, -999, copy=False)
                tmp_dask |= {"DNBRef": (("dim_0", "dim_1"), lunarref_data)}
            except ImportError:
                LOG.info("Failed lunarref in viirs reader.  If you need it, build it")

        # problem with sat_za/az values being too high, need to downsample
        # print("Building xarray")
        obs_xr = xr.Dataset(data_vars=tmp_dask)
        coor_xr = xr.Dataset(data_vars=tmp_coor)
        cal_xr = xr.Dataset(data_vars=tmp_cal_params)

        try:
            tmp_xr = xr.merge([obs_xr, coor_xr, cal_xr])
        except ValueError:
            # downsample for certain bands
            tmp_xr = xr.merge([obs_xr, coor_xr])

        tmp_xr.attrs = {
            "source_file_name": base_fnames,
            "start_datetime": tmp_scn.start_time,
            "end_datetime": tmp_scn.end_time,
            "source_name": tmp_attrs["sensor"],
            "platform_name": tmp_attrs["platform_name"],
            "data_provider": "NOAA",
            "sample_distance_km": tmp_attrs["resolution"] / 1e3,
            "interpolation_radius_of_influence": 1000,
        }

        full_xr |= {var: tmp_xr}
    full_xr["METADATA"] = xr.Dataset(attrs=tmp_xr.attrs)

    return full_xr
